@@ -3,10 +3,8 @@
 /*
  * This file is part of mission-control
  *
- * Copyright (C) 2007-2009 Nokia Corporation.
- * Copyright (C) 2009 Collabora Ltd.
- *
- * Contact: Naba Kumar  <naba.kumar@nokia.com>
+ * Copyright © 2007-2011 Nokia Corporation.
+ * Copyright © 2009-2011 Collabora Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -34,8 +32,6 @@
 #include <dbus/dbus-glib-lowlevel.h>
 #include <dbus/dbus.h>
 
-#include <libmcclient/mc-interfaces.h>
-
 #include <telepathy-glib/dbus.h>
 #include <telepathy-glib/enums.h>
 #include <telepathy-glib/svc-generic.h>
@@ -60,6 +56,8 @@
 #include "plugin-account.h"
 #include "plugin-loader.h"
 
+#include "_gen/interfaces.h"
+
 #define PARAM_PREFIX "param-"
 #define WRITE_CONF_DELAY    500
 
@@ -68,15 +66,16 @@
 
 static void account_manager_iface_init (TpSvcAccountManagerClass *iface,
 					gpointer iface_data);
+static void account_manager_hidden_iface_init (
+    McSvcAccountManagerInterfaceHiddenClass *iface,
+    gpointer iface_data);
 static void properties_iface_init (TpSvcDBusPropertiesClass *iface,
 				   gpointer iface_data);
-static void sso_iface_init (McSvcAccountManagerInterfaceSSOClass *iface,
-                            gpointer data);
 
 static void _mcd_account_manager_constructed (GObject *obj);
 
 static const McdDBusProp account_manager_properties[];
-static const McdDBusProp sso_properties[];
+static const McdDBusProp account_manager_hidden_properties[];
 
 static const McdInterfaceData account_manager_interfaces[] = {
     MCD_IMPLEMENT_IFACE (tp_svc_account_manager_get_type,
@@ -85,9 +84,9 @@ static const McdInterfaceData account_manager_interfaces[] = {
     MCD_IMPLEMENT_IFACE (mc_svc_account_manager_interface_query_get_type,
 			 account_manager_query,
 			 MC_IFACE_ACCOUNT_MANAGER_INTERFACE_QUERY),
-    MCD_IMPLEMENT_IFACE (mc_svc_account_manager_interface_sso_get_type,
-             sso,
-             MC_IFACE_ACCOUNT_MANAGER_INTERFACE_SSO),
+    MCD_IMPLEMENT_IFACE (mc_svc_account_manager_interface_hidden_get_type,
+			 account_manager_hidden,
+			 MC_IFACE_ACCOUNT_MANAGER_INTERFACE_HIDDEN),
     { G_TYPE_INVALID, }
 };
 
@@ -157,10 +156,10 @@ static void account_loaded (McdAccount *account,
 
 /* calback chain for asynchronously updates from backends: */
 static void
-async_altered_validity_cb (McdAccount *account, gboolean valid, gpointer data)
+async_altered_validity_cb (McdAccount *account, const GError *invalid_reason, gpointer data)
 {
     DEBUG ("asynchronously altered account %s is %svalid",
-           mcd_account_get_unique_name (account), valid ? "" : "in");
+           mcd_account_get_unique_name (account), (invalid_reason == NULL) ? "" : "in");
 
     g_object_unref (account);
 }
@@ -286,10 +285,10 @@ altered_one_cb (GObject *storage,
 
 /* callbacks for the various stages in an backend-driven account creation */
 static void
-async_created_validity_cb (McdAccount *account, gboolean valid, gpointer data)
+async_created_validity_cb (McdAccount *account, const GError *invalid_reason, gpointer data)
 {
     DEBUG ("asynchronously created account %s is %svalid",
-           mcd_account_get_unique_name (account), valid ? "" : "in");
+           mcd_account_get_unique_name (account), (invalid_reason == NULL) ? "" : "in");
 
     /* safely cached in the accounts hash by now */
     g_object_unref (account);
@@ -441,23 +440,16 @@ deleted_cb (GObject *plugin, const gchar *name, gpointer data)
 
     if (account != NULL)
     {
+        const gchar * object_path = mcd_account_get_object_path (account);
+
         g_object_ref (account);
+        /* this unhooks the account's signal handlers */
         g_hash_table_remove (manager->priv->accounts, name);
+        tp_svc_account_manager_emit_account_removed (manager, object_path);
         mcd_account_delete (account, _mcd_account_delete_cb, NULL);
     }
 }
 
-GQuark
-mcd_account_manager_error_quark (void)
-{
-    static GQuark quark = 0;
-
-    if (quark == 0)
-        quark = g_quark_from_static_string ("mcd-account-manager-error");
-
-    return quark;
-}
-
 static gboolean
 get_account_connection (const gchar *file_contents, const gchar *path,
                         gchar **p_bus_name, gchar **p_account_name)
@@ -508,6 +500,9 @@ recover_connection (McdAccountManager *account_manager, gchar *file_contents,
     GError *error = NULL;
     gboolean ret = FALSE;
 
+    master = mcd_master_get_default ();
+    g_return_val_if_fail (MCD_IS_MASTER (master), FALSE);
+
     object_path = g_strdelimit (g_strdup_printf ("/%s", name), ".", '/');
     if (!get_account_connection (file_contents, object_path,
                                  &bus_name, &account_name))
@@ -521,9 +516,6 @@ recover_connection (McdAccountManager *account_manager, gchar *file_contents,
     DEBUG ("account is %s", mcd_account_get_unique_name (account));
     manager_name = mcd_account_get_manager_name (account);
 
-    master = mcd_master_get_default ();
-    g_return_val_if_fail (MCD_IS_MASTER (master), FALSE);
-
     manager = _mcd_master_lookup_manager (master, manager_name);
     if (G_UNLIKELY (!manager))
     {
@@ -610,9 +602,18 @@ on_account_validity_changed (McdAccount *account, gboolean valid,
     const gchar *object_path;
 
     object_path = mcd_account_get_object_path (account);
-    tp_svc_account_manager_emit_account_validity_changed (account_manager,
-							  object_path,
-							  valid);
+
+    if (_mcd_account_is_hidden (account))
+    {
+        mc_svc_account_manager_interface_hidden_emit_hidden_account_validity_changed (
+            account_manager, object_path, valid);
+    }
+    else
+    {
+        tp_svc_account_manager_emit_account_validity_changed (account_manager,
+                                                              object_path,
+                                                              valid);
+    }
 }
 
 static void
@@ -623,7 +624,17 @@ on_account_removed (McdAccount *account, McdAccountManager *account_manager)
     const gchar *name, *object_path;
 
     object_path = mcd_account_get_object_path (account);
-    tp_svc_account_manager_emit_account_removed (account_manager, object_path);
+
+    if (_mcd_account_is_hidden (account))
+    {
+        mc_svc_account_manager_interface_hidden_emit_hidden_account_removed (
+            account_manager, object_path);
+    }
+    else
+    {
+        tp_svc_account_manager_emit_account_removed (account_manager,
+                                                     object_path);
+    }
 
     name = mcd_account_get_unique_name (account);
     g_hash_table_remove (priv->accounts, name);
@@ -740,20 +751,10 @@ set_new_account_properties (McdAccount *account,
 }
 
 static void
-complete_account_creation_finish (McdAccount *account, gboolean valid,
-                                  gpointer user_data)
+complete_account_creation_finish (McdAccount *account,
+                                  McdCreateAccountData *cad)
 {
-    McdCreateAccountData *cad = (McdCreateAccountData *) user_data;
-    McdAccountManager *account_manager;
-
-    account_manager = cad->account_manager;
-
-    if (!valid)
-    {
-        cad->ok = FALSE;
-        g_set_error (&cad->error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
-                         "The supplied CM parameters were not valid");
-    }
+    McdAccountManager *account_manager = cad->account_manager;
 
     if (!cad->ok)
     {
@@ -771,6 +772,23 @@ complete_account_creation_finish (McdAccount *account, gboolean valid,
     tp_clear_object (&account);
 }
 
+static void
+complete_account_creation_check_validity_cb (McdAccount *account,
+                                             const GError *invalid_reason,
+                                             gpointer user_data)
+{
+    McdCreateAccountData *cad = user_data;
+
+    if (invalid_reason != NULL)
+    {
+        cad->ok = FALSE;
+        g_set_error_literal (&cad->error, invalid_reason->domain,
+            invalid_reason->code, invalid_reason->message);
+    }
+
+    complete_account_creation_finish (account, cad);
+}
+
 static void
 complete_account_creation_set_cb (McdAccount *account, GPtrArray *not_yet,
                                   const GError *set_error, gpointer user_data)
@@ -784,9 +802,8 @@ complete_account_creation_set_cb (McdAccount *account, GPtrArray *not_yet,
     if (set_error != NULL)
     {
         cad->ok = FALSE;
-        g_set_error (&cad->error, MCD_ACCOUNT_MANAGER_ERROR,
-                     MCD_ACCOUNT_MANAGER_ERROR_SET_PARAMETER,
-                     "Failed to set parameter: %s", set_error->message);
+        g_set_error_literal (&cad->error, set_error->domain, set_error->code,
+                             set_error->message);
     }
 
     if (cad->ok && cad->properties != NULL)
@@ -797,17 +814,11 @@ complete_account_creation_set_cb (McdAccount *account, GPtrArray *not_yet,
     if (cad->ok)
     {
         add_account (account_manager, account, G_STRFUNC);
-        mcd_account_check_validity (account, complete_account_creation_finish, cad);
+        mcd_account_check_validity (account, complete_account_creation_check_validity_cb, cad);
     }
     else
     {
-        complete_account_creation_finish (account, TRUE, cad);
-    }
-
-    if (not_yet != NULL)
-    {
-        g_ptr_array_foreach (not_yet, (GFunc) g_free, NULL);
-        g_ptr_array_free (not_yet, TRUE);
+        complete_account_creation_finish (account, cad);
     }
 }
 
@@ -946,149 +957,15 @@ account_manager_iface_init (TpSvcAccountManagerClass *iface,
 }
 
 static void
-sso_get_service_accounts (McSvcAccountManagerInterfaceSSO *iface,
-                          const gchar *service,
-                          DBusGMethodInvocation *context)
+account_manager_hidden_iface_init (
+    McSvcAccountManagerInterfaceHiddenClass *iface,
+    gpointer iface_data)
 {
-    gsize len;
-    McdAccountManager *manager = MCD_ACCOUNT_MANAGER (iface);
-    McdAccountManagerPrivate *priv = manager->priv;
-    McdStorage *storage = MCD_STORAGE (priv->plugin_manager);
-    GStrv accounts = mcd_storage_dup_accounts (storage, &len);
-    GList *srv_accounts = NULL;
-    GPtrArray *paths = g_ptr_array_new ();
-
-
-    if (len > 0 && accounts != NULL)
-    {
-        guint i = 0;
-        gchar *name;
-
-        for (name = accounts[i]; name != NULL; name = accounts[++i])
-        {
-            gchar *id = mcd_storage_dup_string (storage, name, "libacct-uid");
-
-            if (id != NULL)
-            {
-                gchar *supported =
-                  mcd_storage_dup_string (storage, name, "sso-services");
-
-                if (supported != NULL)
-                {
-                    guint j;
-                    GStrv services = g_strsplit (supported, ";", 0);
-
-                    for (j = 0; services[j] != NULL; j++)
-                    {
-                        if (g_str_equal (service, services[j]))
-                        {
-                            McdAccount *a =
-                              g_hash_table_lookup (priv->accounts, name);
-
-                            if (a != NULL)
-                                srv_accounts = g_list_prepend (srv_accounts, a);
-                        }
-                    }
-
-                    g_free (supported);
-                    g_strfreev (services);
-                }
-
-                g_free (id);
-            }
-        }
-    }
-
-    if (srv_accounts != NULL)
-    {
-        GList *account = srv_accounts;
-
-        for (; account != NULL; account = g_list_next (account))
-        {
-            const gchar *path = mcd_account_get_object_path (account->data);
-            g_ptr_array_add (paths, (gpointer) path);
-        }
-
-        g_list_free (srv_accounts);
-    }
-
-    mc_svc_account_manager_interface_sso_return_from_get_service_accounts (
-      context,
-      paths);
-
-    g_ptr_array_unref (paths);
 }
 
 static void
-sso_get_account (McSvcAccountManagerInterfaceSSO *iface,
-                 const guint id,
-                 DBusGMethodInvocation *context)
-{
-    McdAccountManager *manager = MCD_ACCOUNT_MANAGER (iface);
-    McdAccountManagerPrivate *priv = manager->priv;
-    McdStorage *storage = MCD_STORAGE (priv->plugin_manager);
-    GStrv accounts = mcd_storage_dup_accounts (storage, NULL);
-    const gchar *path = NULL;
-
-    if (accounts != NULL)
-    {
-        guint i = 0;
-        gchar *name;
-        McdAccount *account = NULL;
-
-        for (name = accounts[i]; name != NULL; name = accounts[++i])
-        {
-            gchar *str_id =
-              mcd_storage_dup_string (storage, name, "libacct-uid");
-            guint64 sso_id = g_ascii_strtoull (str_id, NULL, 10);
-
-            if (sso_id != 0 && id != 0)
-            {
-                if (id == sso_id)
-                {
-                    account = g_hash_table_lookup (priv->accounts, name);
-                    break;
-                }
-            }
-
-            g_free (str_id);
-        }
-
-        if (account != NULL)
-            path = mcd_account_get_object_path (account);
-    }
-
-    if (path != NULL)
-    {
-        mc_svc_account_manager_interface_sso_return_from_get_account (context,
-                                                                      path);
-    }
-    else
-    {
-        GError *error = g_error_new (TP_TYPE_ERROR,
-                                     TP_ERROR_DOES_NOT_EXIST,
-                                     "SSO ID %u Not Found", id);
-
-        dbus_g_method_return_error (context, error);
-
-        g_error_free (error);
-    }
-
-    g_strfreev (accounts);
-}
-
-static void
-sso_iface_init (McSvcAccountManagerInterfaceSSOClass *iface, gpointer data)
-{
-#define IMPLEMENT(x) \
-mc_svc_account_manager_interface_sso_implement_##x (iface, sso_##x)
-    IMPLEMENT (get_account);
-    IMPLEMENT (get_service_accounts);
-#undef IMPLEMENT
-}
-
-static void
-accounts_to_gvalue (GHashTable *accounts, gboolean valid, GValue *value)
+accounts_to_gvalue (GHashTable *accounts, gboolean valid, gboolean hidden,
+                    GValue *value)
 {
     static GType ao_type = G_TYPE_INVALID;
     GPtrArray *account_array;
@@ -1106,9 +983,12 @@ accounts_to_gvalue (GHashTable *accounts, gboolean valid, GValue *value)
 
     while (g_hash_table_iter_next (&iter, &k, (gpointer)&account))
     {
-        if (mcd_account_is_valid (account) == valid)
+        if (mcd_account_is_valid (account) == valid &&
+            _mcd_account_is_hidden (account) == hidden)
+        {
             g_ptr_array_add (account_array,
                              g_strdup (mcd_account_get_object_path (account)));
+        }
     }
 
     g_value_init (value, ao_type);
@@ -1123,7 +1003,7 @@ get_valid_accounts (TpSvcDBusProperties *self, const gchar *name,
     McdAccountManagerPrivate *priv = account_manager->priv;
 
     DEBUG ("called");
-    accounts_to_gvalue (priv->accounts, TRUE, value);
+    accounts_to_gvalue (priv->accounts, TRUE, FALSE, value);
 }
 
 static void
@@ -1134,7 +1014,7 @@ get_invalid_accounts (TpSvcDBusProperties *self, const gchar *name,
     McdAccountManagerPrivate *priv = account_manager->priv;
 
     DEBUG ("called");
-    accounts_to_gvalue (priv->accounts, FALSE, value);
+    accounts_to_gvalue (priv->accounts, FALSE, FALSE, value);
 }
 
 static void
@@ -1150,7 +1030,6 @@ get_supported_account_properties (TpSvcDBusProperties *svc,
         TP_IFACE_ACCOUNT ".ConnectAutomatically",
         TP_IFACE_ACCOUNT ".RequestedPresence",
         TP_IFACE_ACCOUNT_INTERFACE_AVATAR ".Avatar",
-        MC_IFACE_ACCOUNT_INTERFACE_COMPAT ".Profile",
         MC_IFACE_ACCOUNT_INTERFACE_COMPAT ".SecondaryVCardFields",
         MC_IFACE_ACCOUNT_INTERFACE_CONDITIONS ".Condition",
         NULL
@@ -1168,8 +1047,30 @@ static const McdDBusProp account_manager_properties[] = {
     { 0 },
 };
 
-static const McdDBusProp sso_properties[] = {
-    { NULL, NULL, NULL },
+static void
+get_valid_hidden_accounts (TpSvcDBusProperties *self, const gchar *name,
+                           GValue *value)
+{
+    McdAccountManager *account_manager = MCD_ACCOUNT_MANAGER (self);
+    McdAccountManagerPrivate *priv = account_manager->priv;
+
+    accounts_to_gvalue (priv->accounts, TRUE, TRUE, value);
+}
+
+static void
+get_invalid_hidden_accounts (TpSvcDBusProperties *self, const gchar *name,
+                             GValue *value)
+{
+    McdAccountManager *account_manager = MCD_ACCOUNT_MANAGER (self);
+    McdAccountManagerPrivate *priv = account_manager->priv;
+
+    accounts_to_gvalue (priv->accounts, FALSE, TRUE, value);
+}
+
+static const McdDBusProp account_manager_hidden_properties[] = {
+    { "ValidHiddenAccounts", NULL, get_valid_hidden_accounts },
+    { "InvalidHiddenAccounts", NULL, get_invalid_hidden_accounts },
+    { 0 },
 };
 
 static void
@@ -1326,13 +1227,13 @@ register_dbus_service (McdAccountManager *account_manager)
     dbus_connection = TP_PROXY (priv->dbus_daemon)->dbus_connection;
 
     if (!tp_dbus_daemon_request_name (priv->dbus_daemon,
-                                      MC_ACCOUNT_MANAGER_DBUS_SERVICE,
+                                      TP_ACCOUNT_MANAGER_BUS_NAME,
                                       TRUE /* idempotent */, &error))
     {
         /* FIXME: put in proper error handling when MC gains the ability to
          * be the AM or the CD but not both */
-        g_error ("Failed registering '%s' service: %s",
-                 MC_ACCOUNT_MANAGER_DBUS_SERVICE, error->message);
+        g_warning("Failed registering '%s' service: %s",
+                  TP_ACCOUNT_MANAGER_BUS_NAME, error->message);
         g_error_free (error);
         exit (1);
     }
@@ -1340,9 +1241,9 @@ register_dbus_service (McdAccountManager *account_manager)
     priv->dbus_registered = TRUE;
 
     if (G_LIKELY (dbus_connection))
-	dbus_g_connection_register_g_object (dbus_connection,
-					     MC_ACCOUNT_MANAGER_DBUS_OBJECT,
-					     (GObject *)account_manager);
+        dbus_g_connection_register_g_object (dbus_connection,
+                                             TP_ACCOUNT_MANAGER_OBJECT_PATH,
+                                             (GObject *)account_manager);
 }
 
 static void
@@ -1578,7 +1479,7 @@ mcd_account_manager_write_conf_async (McdAccountManager *account_manager,
         gsize n_accounts = 0;
 
         groups = mcd_storage_dup_accounts (storage, &n_accounts);
-        DEBUG ("updating all %" G_GSIZE_FORMAT " accounts)", n_accounts);
+        DEBUG ("updating all %" G_GSIZE_FORMAT " accounts", n_accounts);
 
         mcd_storage_commit (storage, NULL);
 
@@ -1610,14 +1511,14 @@ mcd_account_manager_lookup_account_by_path (McdAccountManager *account_manager,
 {
     McdAccountManagerPrivate *priv = account_manager->priv;
 
-    if (!g_str_has_prefix (object_path, MC_ACCOUNT_DBUS_OBJECT_BASE))
+    if (!g_str_has_prefix (object_path, TP_ACCOUNT_OBJECT_PATH_BASE))
     {
         /* can't possibly be right */
         return NULL;
     }
 
     return g_hash_table_lookup (priv->accounts,
-        object_path + (sizeof (MC_ACCOUNT_DBUS_OBJECT_BASE) - 1));
+        object_path + strlen (TP_ACCOUNT_OBJECT_PATH_BASE));
 }
 
 /*
