@@ -28,10 +28,13 @@
 
 #include <errno.h>
 
-#include <telepathy-glib/telepathy-glib.h>
-#include <telepathy-glib/telepathy-glib-dbus.h>
-
+#include <telepathy-glib/dbus.h>
+#include <telepathy-glib/defs.h>
+#include <telepathy-glib/errors.h>
+#include <telepathy-glib/gtypes.h>
+#include <telepathy-glib/interfaces.h>
 #include <telepathy-glib/proxy-subclass.h>
+#include <telepathy-glib/util.h>
 
 #include "channel-utils.h"
 #include "mcd-channel-priv.h"
@@ -61,7 +64,10 @@ static guint signals[N_SIGNALS] = { 0 };
 
 struct _McdClientProxyPrivate
 {
-    GStrv capability_tokens;
+    TpHandleRepoIface *string_pool;
+    /* Handler.Capabilities, represented as handles taken from
+     * dispatcher->priv->string_pool */
+    TpHandleSet *capability_tokens;
 
     gchar *unique_name;
     guint ready_lock;
@@ -324,8 +330,8 @@ parse_client_filter (GKeyFile *file, const gchar *group)
     return filter;
 }
 
-static void _mcd_client_proxy_set_cap_tokens (McdClientProxy *self,
-                                              GStrv cap_tokens);
+static void _mcd_client_proxy_add_cap_tokens (McdClientProxy *self,
+                                              const gchar * const *cap_tokens);
 static void _mcd_client_proxy_add_interfaces (McdClientProxy *self,
                                               const gchar * const *interfaces);
 
@@ -416,7 +422,8 @@ parse_client_file (McdClientProxy *client,
                                       TP_IFACE_CLIENT_HANDLER ".Capabilities",
                                       NULL,
                                       NULL);
-    _mcd_client_proxy_set_cap_tokens (client, cap_tokens);
+    _mcd_client_proxy_add_cap_tokens (client,
+                                      (const gchar * const *) cap_tokens);
     g_strfreev (cap_tokens);
 }
 
@@ -487,7 +494,7 @@ _mcd_client_proxy_set_filters (McdClientProxy *client,
             client_filters = g_list_prepend (client_filters,
                                              new_channel_class);
         else
-            g_hash_table_unref (new_channel_class);
+            g_hash_table_destroy (new_channel_class);
     }
 
     switch (interface)
@@ -515,11 +522,22 @@ _mcd_client_proxy_set_filters (McdClientProxy *client,
 /* This is NULL-safe for the last argument, for ease of use with
  * tp_asv_get_boxed */
 static void
-_mcd_client_proxy_set_cap_tokens (McdClientProxy *self,
-                                  GStrv cap_tokens)
+_mcd_client_proxy_add_cap_tokens (McdClientProxy *self,
+                                  const gchar * const *cap_tokens)
 {
-    g_strfreev (self->priv->capability_tokens);
-    self->priv->capability_tokens = g_strdupv (cap_tokens);
+    guint i;
+
+    if (cap_tokens == NULL)
+        return;
+
+    for (i = 0; cap_tokens[i] != NULL; i++)
+    {
+        TpHandle handle = tp_handle_ensure (self->priv->string_pool,
+                                            cap_tokens[i], NULL, NULL);
+
+        tp_handle_set_add (self->priv->capability_tokens, handle);
+        tp_handle_unref (self->priv->string_pool, handle);
+    }
 }
 
 static void
@@ -603,7 +621,7 @@ _mcd_client_recover_observer (McdClientProxy *self, TpChannel *channel,
         g_hash_table_new (NULL, NULL));
 
     channels_array = _mcd_tp_channel_details_build_from_tp_chan (channel);
-    conn = tp_channel_get_connection (channel);
+    conn = tp_channel_borrow_connection (channel);
     connection_path = tp_proxy_get_object_path (conn);
 
     DEBUG ("calling ObserveChannels on %s for channel %p",
@@ -616,8 +634,8 @@ _mcd_client_recover_observer (McdClientProxy *self, TpChannel *channel,
         NULL, NULL, NULL, NULL);
 
     _mcd_tp_channel_details_free (channels_array);
-    g_ptr_array_unref (satisfied_requests);
-    g_hash_table_unref (observer_info);
+    g_ptr_array_free (satisfied_requests, TRUE);
+    g_hash_table_destroy (observer_info);
 }
 
 static void
@@ -673,7 +691,7 @@ _mcd_client_proxy_handler_get_all_cb (TpProxy *proxy,
      * any capabilities */
     if (self->priv->unique_name[0] != '\0' || self->priv->activatable)
     {
-        _mcd_client_proxy_set_cap_tokens (self,
+        _mcd_client_proxy_add_cap_tokens (self,
             tp_asv_get_boxed (properties, "Capabilities", G_TYPE_STRV));
         g_signal_emit (self, signals[S_HANDLER_CAPABILITIES_CHANGED], 0);
     }
@@ -1005,7 +1023,16 @@ mcd_client_proxy_dispose (GObject *object)
                                             mcd_client_proxy_unique_name_cb,
                                             self);
 
-    tp_clear_pointer (&self->priv->capability_tokens, g_strfreev);
+    if (self->priv->string_pool != NULL)
+    {
+        if (self->priv->capability_tokens != NULL)
+        {
+            tp_handle_set_destroy (self->priv->capability_tokens);
+            self->priv->capability_tokens = NULL;
+        }
+
+        tp_clear_object (&self->priv->string_pool);
+    }
 
     if (chain_up != NULL)
     {
@@ -1047,7 +1074,8 @@ mcd_client_proxy_constructed (GObject *object)
 
     bus_name = tp_proxy_get_bus_name (self);
 
-    self->priv->capability_tokens = NULL;
+    self->priv->capability_tokens = tp_handle_set_new (
+        self->priv->string_pool);
 
     DEBUG ("%s", bus_name);
 
@@ -1081,6 +1109,11 @@ mcd_client_proxy_set_property (GObject *object,
             self->priv->activatable = g_value_get_boolean (value);
             break;
 
+        case PROP_STRING_POOL:
+            g_assert (self->priv->string_pool == NULL);
+            self->priv->string_pool = g_value_dup_object (value);
+            break;
+
         case PROP_UNIQUE_NAME:
             g_assert (self->priv->unique_name == NULL);
             self->priv->unique_name = g_value_dup_string (value);
@@ -1146,6 +1179,14 @@ _mcd_client_proxy_class_init (McdClientProxyClass *klass)
             "TRUE if this client can be service-activated", FALSE,
             G_PARAM_WRITABLE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS));
 
+    g_object_class_install_property (object_class, PROP_STRING_POOL,
+        g_param_spec_object ("string-pool", "String pool",
+            "TpHandleRepoIface used to intern strings representing capability "
+            "tokens",
+            G_TYPE_OBJECT,
+            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
+            G_PARAM_STATIC_STRINGS));
+
     g_object_class_install_property (object_class, PROP_UNIQUE_NAME,
         g_param_spec_string ("unique-name", "Unique name",
             "The D-Bus unique name of this client, \"\" if not running or "
@@ -1164,7 +1205,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
 
     if (!g_ascii_isalpha (*name_suffix))
     {
-        g_set_error (error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
+        g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
                      "Client names must start with a letter");
         return FALSE;
     }
@@ -1173,7 +1214,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
     {
         if (i > (255 - MC_CLIENT_BUS_NAME_BASE_LEN))
         {
-            g_set_error (error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
+            g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
                          "Client name too long");
         }
 
@@ -1186,7 +1227,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
         {
             if (name_suffix[i-1] == '.')
             {
-                g_set_error (error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
+                g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
                              "Client names must not have a digit or dot "
                              "following a dot");
                 return FALSE;
@@ -1194,7 +1235,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
         }
         else
         {
-            g_set_error (error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
+            g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
                          "Client names must not contain '%c'", name_suffix[i]);
             return FALSE;
         }
@@ -1202,7 +1243,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
 
     if (name_suffix[i-1] == '.')
     {
-        g_set_error (error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
+        g_set_error (error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
                      "Client names must not end with a dot");
         return FALSE;
     }
@@ -1212,6 +1253,7 @@ _mcd_client_check_valid_name (const gchar *name_suffix,
 
 McdClientProxy *
 _mcd_client_proxy_new (TpDBusDaemon *dbus_daemon,
+                       TpHandleRepoIface *string_pool,
                        const gchar *well_known_name,
                        const gchar *unique_name_if_known,
                        gboolean activatable)
@@ -1236,6 +1278,7 @@ _mcd_client_proxy_new (TpDBusDaemon *dbus_daemon,
 
     self = g_object_new (MCD_TYPE_CLIENT_PROXY,
                          "dbus-daemon", dbus_daemon,
+                         "string-pool", string_pool,
                          "object-path", object_path,
                          "bus-name", well_known_name,
                          "unique-name", unique_name_if_known,
@@ -1324,7 +1367,7 @@ mcd_client_proxy_free_client_filters (GList **client_filters)
 
     if (*client_filters != NULL)
     {
-        g_list_foreach (*client_filters, (GFunc) g_hash_table_unref, NULL);
+        g_list_foreach (*client_filters, (GFunc) g_hash_table_destroy, NULL);
         g_list_free (*client_filters);
         *client_filters = NULL;
     }
@@ -1387,18 +1430,15 @@ _mcd_client_proxy_get_delay_approvers (McdClientProxy *self)
 static void
 _mcd_client_proxy_become_incapable (McdClientProxy *self)
 {
-    gboolean handler_was_capable = (self->priv->handler_filters != NULL);
-
-    if (self->priv->capability_tokens != NULL &&
-        self->priv->capability_tokens[0] != NULL)
-    {
-        handler_was_capable = TRUE;
-    }
+    gboolean handler_was_capable = (self->priv->handler_filters != NULL ||
+        tp_handle_set_size (self->priv->capability_tokens) > 0);
 
     _mcd_client_proxy_take_approver_filters (self, NULL);
     _mcd_client_proxy_take_observer_filters (self, NULL);
     _mcd_client_proxy_take_handler_filters (self, NULL);
-    tp_clear_pointer (&self->priv->capability_tokens, g_strfreev);
+    tp_handle_set_destroy (self->priv->capability_tokens);
+    self->priv->capability_tokens = tp_handle_set_new (
+        self->priv->string_pool);
 
     if (handler_was_capable)
     {
@@ -1406,14 +1446,29 @@ _mcd_client_proxy_become_incapable (McdClientProxy *self)
     }
 }
 
+typedef struct {
+    TpHandleRepoIface *repo;
+    GPtrArray *array;
+} TokenAppendContext;
+
+static void
+append_token_to_ptrs (TpHandleSet *unused G_GNUC_UNUSED,
+                      TpHandle handle,
+                      gpointer data)
+{
+    TokenAppendContext *context = data;
+
+    g_ptr_array_add (context->array,
+                     g_strdup (tp_handle_inspect (context->repo, handle)));
+}
+
 GValueArray *
 _mcd_client_proxy_dup_handler_capabilities (McdClientProxy *self)
 {
     GPtrArray *filters;
-    GStrv cap_tokens;
+    GPtrArray *cap_tokens;
     GValueArray *va;
     const GList *list;
-    gchar *empty_strv[] = { NULL };
 
     g_return_val_if_fail (MCD_IS_CLIENT_PROXY (self), NULL);
 
@@ -1431,10 +1486,22 @@ _mcd_client_proxy_dup_handler_capabilities (McdClientProxy *self)
         g_ptr_array_add (filters, copy);
     }
 
-    cap_tokens = self->priv->capability_tokens;
+    if (self->priv->capability_tokens == NULL)
+    {
+        cap_tokens = g_ptr_array_sized_new (1);
+    }
+    else
+    {
+        TokenAppendContext context = { self->priv->string_pool, NULL };
 
-    if (cap_tokens == NULL)
-        cap_tokens = empty_strv;
+        cap_tokens = g_ptr_array_sized_new (
+            tp_handle_set_size (self->priv->capability_tokens) + 1);
+        context.array = cap_tokens;
+        tp_handle_set_foreach (self->priv->capability_tokens,
+                               append_token_to_ptrs, &context);
+    }
+
+    g_ptr_array_add (cap_tokens, NULL);
 
     if (DEBUGGING)
     {
@@ -1443,11 +1510,11 @@ _mcd_client_proxy_dup_handler_capabilities (McdClientProxy *self)
         DEBUG ("%s:", tp_proxy_get_bus_name (self));
 
         DEBUG ("- %u channel filters", filters->len);
-        DEBUG ("- %u capability tokens:", g_strv_length (cap_tokens));
+        DEBUG ("- %u capability tokens:", cap_tokens->len - 1);
 
-        for (i = 0; cap_tokens[i] != NULL; i++)
+        for (i = 0; i < cap_tokens->len - 1; i++)
         {
-            DEBUG ("    %s", cap_tokens[i]);
+            DEBUG ("    %s", (gchar *) g_ptr_array_index (cap_tokens, i));
         }
 
         DEBUG ("-end-");
@@ -1464,30 +1531,27 @@ _mcd_client_proxy_dup_handler_capabilities (McdClientProxy *self)
 
     g_value_set_string (va->values + 0, tp_proxy_get_bus_name (self));
     g_value_take_boxed (va->values + 1, filters);
-    g_value_set_boxed (va->values + 2, cap_tokens);
+    g_value_take_boxed (va->values + 2, g_ptr_array_free (cap_tokens, FALSE));
 
     return va;
 }
 
 /* returns TRUE if the channel matches one property criteria
  */
-static gboolean
-_mcd_client_match_property (GVariant *channel_properties,
+gboolean
+_mcd_client_match_property (GHashTable *channel_properties,
                             gchar *property_name,
                             GValue *filter_value)
 {
     GType filter_type = G_VALUE_TYPE (filter_value);
 
-    g_return_val_if_fail (g_variant_is_of_type (channel_properties,
-            G_VARIANT_TYPE_VARDICT), FALSE);
-
     g_assert (G_IS_VALUE (filter_value));
 
     if (filter_type == G_TYPE_STRING)
     {
         const gchar *string;
 
-        string = tp_vardict_get_string (channel_properties, property_name);
+        string = tp_asv_get_string (channel_properties, property_name);
         if (!string)
             return FALSE;
 
@@ -1498,7 +1562,7 @@ _mcd_client_match_property (GVariant *channel_properties,
     {
         const gchar *path;
 
-        path = tp_vardict_get_object_path (channel_properties, property_name);
+        path = tp_asv_get_object_path (channel_properties, property_name);
         if (!path)
             return FALSE;
 
@@ -1510,7 +1574,7 @@ _mcd_client_match_property (GVariant *channel_properties,
         gboolean valid;
         gboolean b;
 
-        b = tp_vardict_get_boolean (channel_properties, property_name, &valid);
+        b = tp_asv_get_boolean (channel_properties, property_name, &valid);
         if (!valid)
             return FALSE;
 
@@ -1523,7 +1587,7 @@ _mcd_client_match_property (GVariant *channel_properties,
         gboolean valid;
         guint64 i;
 
-        i = tp_vardict_get_uint64 (channel_properties, property_name, &valid);
+        i = tp_asv_get_uint64 (channel_properties, property_name, &valid);
         if (!valid)
             return FALSE;
 
@@ -1540,7 +1604,7 @@ _mcd_client_match_property (GVariant *channel_properties,
         gboolean valid;
         gint64 i;
 
-        i = tp_vardict_get_int64 (channel_properties, property_name, &valid);
+        i = tp_asv_get_int64 (channel_properties, property_name, &valid);
         if (!valid)
             return FALSE;
 
@@ -1562,16 +1626,13 @@ _mcd_client_match_property (GVariant *channel_properties,
  * largest filter that matched)
  */
 guint
-_mcd_client_match_filters (GVariant *channel_properties,
+_mcd_client_match_filters (GHashTable *channel_properties,
                            const GList *filters,
                            gboolean assume_requested)
 {
     const GList *list;
     guint best_quality = 0;
 
-    g_return_val_if_fail (g_variant_is_of_type (channel_properties,
-            G_VARIANT_TYPE_VARDICT), 0);
-
     for (list = filters; list != NULL; list = list->next)
     {
         GHashTable *filter = list->data;
@@ -1649,7 +1710,7 @@ borrow_channel_connection_path (McdChannel *channel)
 
     tp_channel = mcd_channel_get_tp_channel (channel);
     g_return_val_if_fail (tp_channel != NULL, "/");
-    tp_connection = tp_channel_get_connection (tp_channel);
+    tp_connection = tp_channel_borrow_connection (tp_channel);
     g_return_val_if_fail (tp_connection != NULL, "/");
     connection_path = tp_proxy_get_object_path (tp_connection);
     g_return_val_if_fail (connection_path != NULL, "/");
@@ -1725,6 +1786,6 @@ _mcd_client_proxy_handle_channels (McdClientProxy *self,
         callback, user_data, destroy, weak_object);
 
     _mcd_tp_channel_details_free (channel_details);
-    g_ptr_array_unref (requests_satisfied);
+    g_ptr_array_free (requests_satisfied, TRUE);
     g_hash_table_unref (handler_info);
 }
